"""apetype module

Provides base class `Settings` that can be used
for making a derivative class that contains typed
settings, for a module or package.
"""

from .configs import ConfigBase
from .tasks import TaskBase
